// BATHIUM CANADA CONFIDENTIAL
//-----------------------------------------------------------------------------
//
// [2013] Bathium Canada
// All Rights Reserved.
//
// NOTICE:  All information contained herein is, and remains
// the property of Bathium Canada Incorporated
// The intellectual and technical concepts contained
// herein are proprietary to Bathium Canada and may be covered by
// Canadian, U.S. and/or  Foreign Patents,
// patents in process, and are protected by trade secret or copyright law.
// Dissemination of this information or reproduction of this material
// is strictly forbidden unless prior written permission is obtained
// from Bathium Canada.
//-----------------------------------------------------------------------------
#include "tsCAN.h"
#include "ts/libtsctl.c"
#include <string.h>
#include "ts/NetCANctl.c"
#include "ts/Net.c"
#include "ts/Array.h"
// TYPEDEFS

typedef struct _tsCANHandle
{
  CAN * Handle;
  CANPort can;
  HWComInterface interface;
}tsCANHandle;

// PROTOTYPES
STATIC CAN * Net_CANctlInit( s32 inst,Character_t *host, s32 mode);
STATIC BatErr_t FindComInterfaceInstance( struct _HWComInterface * me, u32 * p_Instance );

// CAN Interface
STATIC BatErr_t tsCAN_Init( struct _HWComInterface * me );
STATIC BatErr_t tsCAN_Configure( struct _HWComInterface * me, void * );
STATIC BatErr_t tsCAN_Control( struct _HWComInterface * me, HWComControlCommand * );
STATIC BatErr_t tsCAN_Enable( struct _HWComInterface * me );
STATIC BatErr_t tsCAN_Disable( struct _HWComInterface * me );
STATIC BatErr_t tsCAN_Transmit( struct _HWComInterface * me, HWComInterfaceMessage * );
STATIC BatErr_t tsCAN_Receive( struct _HWComInterface * me, HWComInterfaceMessage * );

// PRIVATE VARS

STATIC tsCANHandle g_tsCANHandle[MAX_TS_CAN_CHANNELS];

// PRIVATE FCT

STATIC CAN * Net_CANctlInit( s32 inst,Character_t *host, s32 mode)
{
  s32 socket = ClientSocketNew(host,7552+inst);
  if( socket < 0 ) return 0;
  NetCANctl *obj = malloc(sizeof(NetCANctl));
  memset(obj,0,sizeof(*obj));
  return NetCANctlInit(obj,socket);
}

STATIC BatErr_t FindComInterfaceInstance( struct _HWComInterface * me, u32 * p_Instance )
{
  u32 i = 0;
  BatErr_t l_Result = ERROR;

  if( ( me != NULL ) && ( p_Instance != NULL ) )
  {
    while( ( i < MAX_TS_CAN_CHANNELS ) && ( l_Result == ERROR ) )
    {
      if( me == &g_tsCANHandle[i].interface )
      {
        *p_Instance = i;
        l_Result = SUCCESS;
      }
      ++i;
    }
  }

  return( l_Result );
}

// CAN Interface
STATIC BatErr_t tsCAN_Init( struct _HWComInterface * me )
{
  BatErr_t l_Result = ERROR;
  u32 l_Instance;
  u32 l_RetryNb = 100;

  if( FindComInterfaceInstance( me, &l_Instance ) == SUCCESS )
  {
    ThreadInit( );
    g_tsCANHandle[ l_Instance ].can.can  = CANInit( l_Instance );
    g_tsCANHandle[ l_Instance ].can.port = ( 7552 + l_Instance );
    ThreadNew( "canctl server", l_Instance, CANStart, 0, &g_tsCANHandle[ l_Instance ].can, 0 );

    g_tsCANHandle[ l_Instance ].Handle = NULL;

    while( ( l_RetryNb-- ) && ( g_tsCANHandle[ l_Instance ].Handle == NULL ) )
    {
      g_tsCANHandle[ l_Instance ].Handle = Net_CANctlInit( l_Instance, "127.0.0.1", 0 ); 
    }
    if( g_tsCANHandle[ l_Instance ].Handle != NULL )
    {
      l_Result = SUCCESS;
    }
  }

  return( l_Result );
}
STATIC BatErr_t tsCAN_Configure( struct _HWComInterface * me, void * p_Config )
{
  BatErr_t l_Result = ERROR;
  tsCANConfig * l_tsCanConfig = ( tsCANConfig * )p_Config;
  u32 l_Instance;

  if( FindComInterfaceInstance( me, &l_Instance ) == SUCCESS )
  {
    if( l_tsCanConfig != NULL )
    {
      g_tsCANHandle[l_Instance].Handle->BaudSet( g_tsCANHandle[l_Instance].Handle, l_tsCanConfig->BitRate );
      l_Result = SUCCESS;
    }
  }

  return( l_Result );
}
STATIC BatErr_t tsCAN_Control( struct _HWComInterface * me, HWComControlCommand * p_Control )
{
  BatErr_t l_Result = ERROR;

  return( l_Result );
}
STATIC BatErr_t tsCAN_Enable( struct _HWComInterface * me )
{
  BatErr_t l_Result = SUCCESS;

  return( l_Result );
}
STATIC BatErr_t tsCAN_Disable( struct _HWComInterface * me )
{
  BatErr_t l_Result = SUCCESS;

  return( l_Result );
}
STATIC BatErr_t tsCAN_Transmit( struct _HWComInterface * me, HWComInterfaceMessage * p_Msg )
{
  BatErr_t l_Result = ERROR;
  u32 l_Instance;

  if( FindComInterfaceInstance( me, &l_Instance ) == SUCCESS )
  {
    if( ( p_Msg != NULL ) && ( p_Msg->Msg != NULL ) )
    {
      ArrayAuto(char,data,ARR(p_Msg->Msg[1],p_Msg->Msg[2],p_Msg->Msg[3],p_Msg->Msg[4],p_Msg->Msg[5],p_Msg->Msg[6],p_Msg->Msg[7]));

      if( g_tsCANHandle[l_Instance].Handle->Tx( g_tsCANHandle[l_Instance].Handle,0 , p_Msg->MsgID, (const char *)data ) == CANSuccess )
      {
        printf("Tx Success\n");
        l_Result = SUCCESS;
      }
    }
  }

  return( l_Result );
}
STATIC BatErr_t tsCAN_Receive( struct _HWComInterface * me, HWComInterfaceMessage * p_Msg )
{
  BatErr_t l_Result;
  CANMessage l_Msg;
  u32 l_Instance;

  if( FindComInterfaceInstance( me, &l_Instance ) == SUCCESS )
  {
    if( ( p_Msg != NULL ) && ( p_Msg->Msg != NULL ) )
    {
      printf("Receiving...\n");
      if( g_tsCANHandle[l_Instance].Handle->Rx( g_tsCANHandle[l_Instance].Handle, &l_Msg ) == CANSuccess )
      {
        printf("got something!\n");
        p_Msg->MsgID = l_Msg.id;
        p_Msg->MsgSize = l_Msg.length;
        memcpy( p_Msg->Msg, l_Msg.data, l_Msg.length );
        l_Result = SUCCESS;
      }
    }
  }

  return( l_Result );
}

// PUBLIC FCT

HWComInterface * tsCAN_GetInterface( tsCANChannel p_Channel )
{
  HWComInterface * l_Interface = NULL;

  if( p_Channel < MAX_TS_CAN_CHANNELS )
  {
     g_tsCANHandle[p_Channel].interface.Initialise = tsCAN_Init;
     g_tsCANHandle[p_Channel].interface.Configure  = tsCAN_Configure;
     g_tsCANHandle[p_Channel].interface.Enable     = tsCAN_Enable;
     g_tsCANHandle[p_Channel].interface.Disable    = tsCAN_Disable;
     g_tsCANHandle[p_Channel].interface.Control    = tsCAN_Control;
     g_tsCANHandle[p_Channel].interface.Transmit   = tsCAN_Transmit;
     g_tsCANHandle[p_Channel].interface.Receive    = tsCAN_Receive;

    l_Interface = &g_tsCANHandle[ p_Channel ].interface;
  }

  return( l_Interface );
}

